#define Titel "CCForthF"
/* A Small Forth
 *  
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define pico
//#define P2

#ifdef pico
#include <LittleFS.h>
#endif

#ifdef P2
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/vfs.h>

enum { _xtlfreq = 25_000_000 } ;
enum {    _clkfreq = 200_000_000 } ;
enum {    DOWNLOAD_BAUD = 921_600 } ; // 230_400
enum {    DEBUG_BAUD = 921_600 } ;
enum { HEAPSIZE= 4*1024 };
#endif

//#include <propeller.h>

#ifndef PATH_MAX
#define PATH_MAX 256
#endif
#define BSIZE 32768

#ifdef pico
#define FORTHSIZE 0x30000 // 0x30000= 196608 300000
int32_t TOS, NOS, SP, RP, auxP, IP, W, EABORT, UStart, status, 
   toset, trace, R0, R1, R2, R3; // 
char* forthMem; // pico: malloc used
#define printf(...) Serial.printf(__VA_ARGS__)
#endif

#ifdef P2
#define int32_t int
#define FORTHSIZE 300000 // 0x30000= 196608 300000
extern register int32_t TOS, NOS, SP, RP, auxP, IP, W, EABORT, UStart, status, 
   toset, trace, R0, R1, R2, R3; 
char forthMem[FORTHSIZE]={0};
unsigned char cstack[8][128]; // stacks for C in other COGs
#endif
  
int32_t TIB, wordBuf, inP, DictStart, CP, CONTEXT, LAST, primLimit, LATEST;
int32_t lastforth, factBlock=0, flineCnt=0, doesP, background=0;

// macros

void codes____________dumy() {} // for function list
// codes ##########################################################################
enum { endword, noop, wexit, prtS4, dump16, drop, twodrop, push, dup, twodup,// 
   store, at, hereat, emit, dot, plus, minus, lit, // 
   sttora, ratost, cpratost, swap, rot, over, // 
   lwords, words, findstr, tick, execute, prtword, see, nip, 
   comma, ccomma, colon, semicolon, dofnext, ffor, fnext, evaluate, dovar, 
   variable, docreate, create, doconst, constant, dovalue, value, to, plusto,
   toLocA, plusToLocA, getLocA, toLocB, plusToLocB, getLocB,
   toLocC, plusToLocC, getLocC, toLocD, plusToLocD, getLocD,  
   toLocE, plusToLocE, getLocE, 
   loc, endloc, loc_a, loc_ab, loc_abc, loc_abcd, loc_abcde,
   backslash, doparen, allot, mult, fdiv,
   muldiv, plus1, minus1, plusstore, branch, nullbranch, immediate, 
   equal, less, more, iszero, negate, begin, until, fif, fthen, felse,
   again, fwhile, repeat, fdo, doloop, floop, doplusloop, plusloop,
   i, j, ftor, ftand, ftxor, ftmin, ftmax, mod, cstore, cat, fill,
   getstring, defstring, prints, prtstring, lastfree,
   freadBlock, fwriteBlock, dir, format, bSize, bBuf, actBlock, lineCnt,
   conkey, gettext, forget, context, finP, ftword, interpret, compile,
   depth, rdepth, shiftl, shiftr, fbackground, pause, getUser, settrace,
   cmove, rcmove,
    // processor specific
    cntat, bye, pinat, pinstore, fgetus, fgetms,
    cogstart
};

// flags for words #########################################################################
#define IMMED 128
#define CONLY 64
enum { ISNORMAL, ISLOOP, ISSTRUCT, ISDATA, ISIO, ISASS };
#define PRIMI(name) const char* name##_s = #name; \
   void _##name(void)
#define PRIMS(name,string) const char* name##_s = #string; \
   void _##name(void)

#ifdef P2 
// buffered input via extra COG1 ###########################################################
// 2^n for auto wrap around!!!

#define INBUFSIZE 4095
char inBuf[INBUFSIZE+1];
char lfflag=0, // line end received
   cCflag=0; // cntrl C received
   
int32_t wrPoint=0, rdPoint=0;
unsigned char bistack[128];

void bufferedInput(void) {
   char inChar;
   while(1) {
      inChar=_rxraw(0);
      if(inChar==10) lfflag++; // got line end
      if(inChar==3) cCflag++; // got cntrl C
      inBuf[wrPoint++]=inChar;
      wrPoint&=INBUFSIZE; // wrap around
   }
}

PRIMI(conkey) { // ( -- char ) 0 for none
   if(wrPoint==rdPoint) { // no key in buffer
      // background??
      _pushx(0);
      return;
   }
   R0=(int32_t)inBuf[rdPoint++];
   rdPoint&=INBUFSIZE; // wrap around
   if(R0==10 && lfflag>0) lfflag--;
   _pushx(R0);
}

void bufferedgets(char tib[]) { // get a line
   unsigned char tipoint=0;
   tib[tipoint]=0;
   do{
      while(rdPoint==wrPoint) { // no char received
         if(background!=0) { // do a background task
            W=background;
            if(W<=primLimit+29) W=*(int32_t*)W;
            else W=-W;
            inner();
         }
      }
      R0=(int32_t)inBuf[rdPoint++];
      rdPoint&=INBUFSIZE; // wrap around
      if(R0==10 && lfflag>0) lfflag--;
      if(R0==8 && tipoint>0) tipoint--;
      if(R0>31 || R0== 8 || R0==10 || R0==13) { 
         // printf("%d %d ",wrPoint, rdPoint);
         // _txraw('R'); 
         // printf(" %d ",R0);
         _txraw(R0); }
      if(R0>31 || R0==10 || R0==13) { 
         tib[tipoint++]=R0;
         tib[tipoint]=0;
      }
   } while(R0!=10 && R0!=13); // line end
   tib[tipoint]=0;
}

PRIMI(gettext) { // get text into block Buffer until ESC=27 or EURO=128
   R1= (int32_t)(forthMem+FORTHSIZE-BSIZE);  
   do {
      while(rdPoint==wrPoint); // no char received
      R0=(int32_t)inBuf[rdPoint++];
      rdPoint&=INBUFSIZE; // wrap around
      if(R0==10 && lfflag>0) lfflag--; // line feed
      if(R0!=27 ) { *(char*)R1 = R0; R1++; } 
      if(R0==10) putchar('.'); // print dot for LF
   } while(R0!=27 && R0!=128);
   *(char*)R1 = 0; 
}
#endif

#ifdef pico
// for pico ##############################################################
char lfflag=0, // line end received
   cCflag=0; // cntrl C received
   
PRIMI(conkey){ // ( -- char ) 0 for none
   if(Serial.available()>0) {
      _pushx(Serial.read());   
   }
   else _pushx(0); 
}

void bufferedgets(char tib[]) { // get a line
   unsigned char tipoint=0;
   tib[tipoint]=0;
   do{
      while(Serial.available()==0) { // no char received
         if(background!=0) { // do a background task
            W=background;
            if(W<=primLimit+29) W=*(int32_t*)W;
            else W=-W;
            inner();
         }
      }
      R0=Serial.read();
      if(R0==10 && lfflag>0) lfflag--;
      if(R0==8 && tipoint>0) tipoint--;
      if(R0>31 || R0== 8 || R0==10 || R0==13) { 
         Serial.write(R0);
         }
      if(R0>31 || R0==10 || R0==13) { 
         tib[tipoint++]=R0;
         tib[tipoint]=0;
      }
   } while(R0!=10 && R0!=13); // line end
   tib[tipoint]=0;
}

PRIMI(gettext) { // get text into block Buffer until ESC=27 or EURO=128
   R1= (int32_t)(forthMem+FORTHSIZE-BSIZE); 
   do {
	   while(Serial.available()==0); // no char received
	   R0=Serial.read();
      if(R0!=27 ) { *(char*)R1 = R0; R1++; } 
      if(R0==10) Serial.write('.'); // print dot for LF
   } while(R0!=27 && R0!=128);
   *(char*)R1 = 0; 
}
#endif


// Primitive words #####################################################################
const char* noop_s="nop";
void _noop(void)
{ }

const char* prtS4_s= "prtS4";
void _prtS4(void)
{
   printf("\n $%x $%x $%x $%x ", *(int32_t*)(SP-4), *(int32_t*)SP, NOS, TOS);
   printf("\n %d %d %d %d > ", *(int32_t*)(SP-4), *(int32_t*)SP, NOS, TOS);
}

const char* drop_s="drop";
void _drop(void) { TOS= NOS; NOS= *(int32_t*)SP; if(SP>0)SP-=4; }
// #define _drop() TOS= NOS; NOS= *(int32_t*)SP; if(SP>0)SP-=4 // ist langsamer!

const char* dump16_s="dump16"; // dump 16 bytes from TOS
void _dump16( void )
{
   printf("\n %8x ",TOS);
   for(int32_t i=0;i<16;i++) {
      printf(" %2x", (char)*(char*)TOS++); 
      if(i==3||i==7||i==11) printf(",");
   }
   _drop();
}

const char* twodrop_s="2drop";
void _twodrop(void) 
{ TOS= *(int32_t*)SP; NOS= *(int32_t*)(SP-4); if(SP>4)SP-=8; }

// no forth word
void _pushx(int32_t x)
{
   SP+=4;
   *(int32_t*)SP=NOS;
   NOS=TOS;
   TOS=x;
}

const char* dup_s="dup";
void _dup(void)
{
   _pushx(TOS);
}             

const char* twodup_s="2dup";
void _twodup(void)
{
   R1=TOS;
   _pushx(NOS);
   _pushx(R1);
}

const char* store_s="!";
void _store(void) 
{
   *(int32_t*)TOS= NOS;
   _twodrop();
}

const char* at_s="@";
void _at( void)
{
   TOS= *(int32_t*)TOS;
}

const char* hereat_s="here@";
void _hereat( void)
{
   _pushx(CP);
}

#ifdef P2
const char* emit_s="emit";
void _emit( void)
{
   putchar((char)TOS);
   _drop();
}
#endif

#ifdef pico
const char* emit_s="emit";
void _emit( void)
{
   Serial.write((char)TOS);
   _drop();
}
#endif

const char* dot_s=".";
void _dot( void)
{
   printf(" %d", TOS);
   _drop();
}

const char* plus_s="+";
void _plus( void)
{
   NOS+= TOS;
   _drop();
}

const char* minus_s="-";
void _minus( void)
{
   NOS-= TOS;
   _drop();
}

const char* wexit_s="exit";
void _wexit(void)
{
   IP=*(int32_t*)RP;
   RP-=4;
}

const char* lit_s="lit";
void _lit(void)
{
   _pushx(*(int32_t*)IP);
   IP+=4;
}

// no forth word
void _rpushx(int32_t x)
{
   RP+=4;
   *(int32_t*)RP=x;
}


const char* sttora_s= ">r";
void _sttora(void) 
{ // stack to rack
   _rpushx(TOS);
   _drop();
}

const char* ratost_s="r>";
void _ratost(void) 
{  // rack to stack
   _pushx(*(int32_t*)RP);
   RP-=4;
}

const char* cpratost_s="r@";
void _cpratosta(void) 
{  // copy rack to stack
   _pushx(*(int32_t*)RP);
}

const char* swap_s="swap";
void _swap(void) 
{  
   R0=TOS;
   TOS=NOS;
   NOS=R0;
}

const char* rot_s="rot";
void _rot(void) 
{  
   R0=*(int32_t*)SP;
   *(int32_t*)SP=NOS;
   NOS=TOS;
   TOS=R0;
}

const char* over_s="over";
void _over(void)
{
   _pushx(NOS);
}

const char* lwords_s="lwords"; // mit link field
void _lwords(void)
{
   int32_t i=LAST;
   do{
      printf("%-20s %2x %8.4x \n", (char*)(i+12), *(char*)(i+8), i);
      i=*(int32_t*)i;
   } while(i!=0);
}

const char* words_s="words";
void _words(void)
{
   int32_t i=LAST;
   do{
      printf("%s ", (char*)(i+12));
      i=*(int32_t*)i;
   } while(i!=0);
   printf("\n");
}

int32_t _ffindstr(char* c) { // gets linkfield from Name, 0 if not found
   int32_t i=LAST;
   do{
      if(strcmp((char*)(i+12), c)==0) return i;
      i=*(int32_t*)i;
   } while(i!=0);
   return 0;
}

const char* findstr_s="find$"; // linkfield
void _findstr(void)
{
   TOS=_ffindstr((char*)TOS);
}

const char* tick_s="'"; // get codefield from wordname
void _tick(void)
{
   fword();
   R0=_ffindstr((char*)wordBuf);
   if(R0!=0) _pushx(32+R0);
   else _pushx(0);
}


const char* execute_s="execute"; // execute codefieldaddress from TOS
void _execute(void)
{
   W=TOS;
   _drop();
   if(W<=primLimit+32) W=*(int32_t*)W;
   else W=-W;
   inner();
}


// not forth
void fprtWord(int32_t code) { // print word name from it's code
   code=abs(code);
   if(code<=primLimit) {
      int32_t i=primLimit;
      do{
         if(*(int32_t*)(i+32)==code) { 
            printf(" %s \n", (char*)(i+12));
            return;
         }
         i=*(int32_t*)i;
      } while(i!=0);
   }

   int32_t i=LAST;
   do{
      if(i==code-32) { 
         printf("  %s \n", (char*)(i+12));
         return;
      }
      i=*(int32_t*)i;
   } while(i!=0);
   printf("-- Not found \n");
}

PRIMI(prtword) { // ( code -- ) from code
   fprtWord(TOS);
   _drop();
}

// not forth
void fseeLinkfield(int32_t linkfield) {
   printf("%s: %8x \n", (char*)(linkfield+12), linkfield); 
   R0=linkfield+32-4;
   do{
      R0+=4;
      printf(" %8x %8x ", R0, *(int32_t*)R0);
      fprtWord(abs(*(int32_t*)R0));
   if((*(int32_t*)R0)==lit || (*(int32_t*)R0)==branch || (*(int32_t*)R0)==nullbranch || (*(int32_t*)R0)==doloop \
   || (*(int32_t*)R0)==doplusloop ) { 
      printf("        %16d \n", *(int32_t*)(R0+4)); R0+=4; }
   }while(*(int32_t*)R0!=wexit && *(int32_t*)R0!=getstring);
}


// not forth
void fsee(char* word) {
   int32_t co= _ffindstr(word); // points to linkfield
   if(co==0) {
      printf(" -not found\n");
      return;
   }
   if(co<=primLimit) {
      printf(" is Primitive\n");
      return;
   }
   fseeLinkfield(co);
}

const char* see_s="see"; // see <word> decompile word
void _see(void) {
   fword();
   fsee((char*)wordBuf);
}

PRIMI(nip) {
   NOS=TOS;
   _drop();
}

const char* comma_s=",";
void _comma(void) {
   *(int32_t*)CP=TOS;
   _drop();
   CP+=4;
}

const char* ccomma_s="c,";
void _ccomma(void) {
   *(char*)CP=TOS;
   _drop();
   CP++;
}

const char* colon_s=":";
void _colon(void) {
	while((CP&3)!=0) CP++; // ALIGNEMENT
   LATEST=CP;
   *(int32_t*)CP=LAST; // link field
   CP+=4; //
   *(int32_t*)CP=0; // source field
   CP+=4;
   *(char*)CP=0; // flags field
   CP+=4;
   fword(); // get name into wordBuf
   printf("%s",(char*)wordBuf);
   for(R0=0;R0<20;R0+=4){ // copy name into namefield
      *(int32_t*)CP=*(int32_t*)(wordBuf+R0);
      CP+=4;
   }
   status=1; // switch to compile
}

const char* semicolon_s=";";
void _semicolon(void) {
   *(int32_t*)CP=wexit; CP+=4;
   LAST=LATEST;
   CONTEXT=LATEST;
   status=0; // switch to interprete
}

PRIMI(dofnext) { // fast next down counter top on rack
   (*(int32_t*)RP)--; // decrement top of returnstack
   if(*(int32_t*)RP != 0)
   {
      IP=(*(int32_t*)IP); // jump 
   }
   else
   {
      RP-=4; // drop counter from rack
      IP+=4; // jump over destination
   }
}

PRIMI(ffor) { // immediate
   *(int32_t*)CP=sttora; // get count onto rack
   CP+=4;
   _pushx(CP);
}

PRIMI(fnext) { // immediate
   *(int32_t*)CP=dofnext; // 
   CP+=4;
   *(int32_t*)CP=TOS; // compile jump address
   _drop();
   CP+=4;
}

PRIMI(evaluate) { // a stringaddr on TOS
   inP=TOS; _drop();
   while(*(char*)inP!=0) {
      fevalWord();
      //printf("\n RDepth: %d Depth: %d ", (RP-(int32_t)Rack)/4, (SP-(int32_t)Stack)/4);
      //_prtS4();
   }
}

PRIMI(dovar) { // get address
   _pushx((IP+4)); // next will be exit, data in next to next
}

PRIMI(variable) { // creates header and space for variable
   while((CP&3)!=0) CP++; // ALIGNEMENT
   LATEST=CP; // preserve link field address
   *(int32_t*)CP=LAST; // link field
   CP+=4; //
   *(int32_t*)CP=0; // source field
   CP+=4;
   *(char*)CP=ISDATA; // flags field byte
   CP+=4;
   fword(); // get name into wordBuf
   printf("%s",(char*)wordBuf);
   for(R0=0;R0<20;R0+=4){ // copy name into namefield
      *(int32_t*)CP=*(int32_t*)(wordBuf+R0);
      CP+=4;
   }
   *(int32_t*)CP=dovar; CP+=4; // 
   *(int32_t*)CP=wexit; 
   CP+=8; // reserve data space
   LAST=LATEST;
   CONTEXT=LATEST;
}

PRIMI(docreate) { // get address
   _pushx((IP+8)); // next will be does, then exit; data in 3rd, if reserved
}

PRIMI(create) { // creates header but no space for variable
	while((CP&3)!=0) CP++; // ALIGNEMENT
   LATEST=CP; // preserve link field address
   *(int32_t*)CP=LAST; // link field
   CP+=4; //
   *(int32_t*)CP=0; // source field
   CP+=4;
   *(char*)CP=ISDATA; // flags field byte
   CP+=4;
   fword(); // get name into wordBuf
   printf("%s",(char*)wordBuf);
   for(R0=0;R0<20;R0+=4){ // copy name into namefield
      *(int32_t*)CP=*(int32_t*)(wordBuf+R0);
      CP+=4;
   }
   *(int32_t*)CP=docreate; CP+=4; // will get address of data
   *(int32_t*)CP=wexit; doesP = CP; // reserve for patch by does>
   CP+=4;
   *(int32_t*)CP=wexit; 
   CP+=4; // here can be data space, not reserved
   LAST=LATEST;
   CONTEXT=LATEST;
}

PRIMI(doconst) { // get data
   _pushx(*(int32_t*)(IP+4)); // next will be exit, data in next to next
}

PRIMI(constant) { // creates header and space for constant, stores stack
	while((CP&3)!=0) CP++; // ALIGNEMENT
   LATEST=CP; // preserve link field address
   *(int32_t*)CP=LAST; // link field
   CP+=4; //
   *(int32_t*)CP=0; // source field
   CP+=4;
   *(char*)CP=ISDATA; // flags field byte
   CP+=4;
   fword(); // get name into wordBuf
   printf("%s",(char*)wordBuf);
   for(R0=0;R0<20;R0+=4){ // copy name into namefield
      *(int32_t*)CP=*(int32_t*)(wordBuf+R0);
      CP+=4;
   }
   *(int32_t*)CP=doconst; CP+=4; // 
   *(int32_t*)CP=wexit; CP+=4;
   *(int32_t*)CP=TOS; CP+=4; // put data
   _drop();
   LAST=LATEST;
   CONTEXT=LATEST;
}

PRIMI(dovalue) { // handle value type variables
   if(toset==0) // get data
      _pushx(*(int32_t*)(IP+4)); // next will be exit, data in next to next
   else { // set data 
      if(toset==1) 
         *(int32_t*)(IP+4)=TOS; 
      
      if(toset==2) // add data
         *(int32_t*)(IP+4) += TOS; 
      
      _drop(); 
      toset=0;
   }
}

PRIMI(value) { // creates header and space for value type variable, stores stack
	while((CP&3)!=0) CP++; // ALIGNEMENT
   LATEST=CP; // preserve link field address
   *(int32_t*)CP=LAST; // link field
   CP+=4; //
   *(int32_t*)CP=0; // source field
   CP+=4;
   *(char*)CP=ISDATA; // flags field byte
   CP+=4;
   fword(); // get name into wordBuf
   printf("%s",(char*)wordBuf);
   for(R0=0;R0<20;R0+=4){ // copy name into namefield
      *(int32_t*)CP=*(int32_t*)(wordBuf+R0);
      CP+=4;
   }
   *(int32_t*)CP=dovalue; CP+=4; // 
   *(int32_t*)CP=wexit; CP+=4;
   *(int32_t*)CP=TOS; CP+=4; // put data
   _drop();
   LAST=LATEST;
   CONTEXT=LATEST;
}

PRIMI(to) {
   toset=1;
}

PRIMS(plusto,+to) {
   toset=2;
}


PRIMS(toLocA,>a) {
   *(int32_t*)auxP=TOS;
   _drop();
}

PRIMS(plusToLocA,+>a) {
   *(int32_t*)auxP += TOS;
   _drop();
}

PRIMS(getLocA,a>) {
   _pushx(*(int32_t*)auxP);
}

PRIMS(toLocB,>b) { *(int32_t*)(auxP+4)=TOS; _drop(); }
PRIMS(plusToLocB,+>b) { *(int32_t*)(auxP+4) += TOS; _drop(); }
PRIMS(getLocB,b>) { _pushx(*(int32_t*)(auxP+4)); }


PRIMS(toLocC,>c) { *(int32_t*)(auxP+8)=TOS; _drop(); }
PRIMS(plusToLocC,+>c) { *(int32_t*)(auxP+8) += TOS; _drop(); }
PRIMS(getLocC,c>) { _pushx(*(int32_t*)(auxP+8)); }

PRIMS(toLocD,>d) { *(int32_t*)(auxP+12)=TOS; _drop(); }
PRIMS(plusToLocD,+>d) { *(int32_t*)(auxP+12) += TOS; _drop(); }
PRIMS(getLocD,d>) { _pushx(*(int32_t*)(auxP+12)); }

PRIMS(toLocE,>e) { *(int32_t*)(auxP+16)=TOS; _drop(); }
PRIMS(plusToLocE,+>e) { *(int32_t*)(auxP+16) += TOS; _drop(); }
PRIMS(getLocE,e>) { _pushx(*(int32_t*)(auxP+16)); }

PRIMI(loc) {
   auxP+=20; // 5*4
}

PRIMI(loc_a) {
   auxP+=20; // 5*4
   *(int32_t*)(auxP)=TOS; _drop();
}
PRIMI(loc_ab) {
   auxP+=20; // 5*4
   *(int32_t*)(auxP+4)=TOS; _drop();
   *(int32_t*)(auxP+0)=TOS; _drop();
}
PRIMI(loc_abc) {
   auxP+=20; // 5*4
   *(int32_t*)(auxP+8)=TOS; _drop();
   *(int32_t*)(auxP+4)=TOS; _drop();
   *(int32_t*)(auxP+0)=TOS; _drop();
}
PRIMI(loc_abcd) {
   auxP+=20; // 5*4
   *(int32_t*)(auxP+12)=TOS; _drop();
   *(int32_t*)(auxP+8)=TOS; _drop();
   *(int32_t*)(auxP+4)=TOS; _drop();
   *(int32_t*)(auxP+0)=TOS; _drop();
}
PRIMI(loc_abcde) {
   auxP+=20; // 5*4
   *(int32_t*)(auxP+16)=TOS; _drop();
   *(int32_t*)(auxP+12)=TOS; _drop();
   *(int32_t*)(auxP+8)=TOS; _drop();
   *(int32_t*)(auxP+4)=TOS; _drop();
   *(int32_t*)(auxP+0)=TOS; _drop();
}

PRIMI(endloc) {
   auxP-=20; // 5*4
}

const char* backslash_s="\\";
void _backslash(void) { // comment until LF
   while((*(char*)inP != 0xA)&&(*(char*)inP != 0)) inP++;
   if((*(char*)inP != 0)) inP++;
}

const char* doparen_s="(";
void _doparen(void) { // comment until )
   while((*(char*)inP != 0x29)&&(*(char*)inP != 0)) inP++;
   inP++;
}

PRIMI(allot) { 
   CP+=TOS;
   while((CP&3)!=0) { // ALIGNEMENT
	   CP++;
   }
   _drop();
}

PRIMS(mult,\x2A ) {
   NOS*= TOS;
   _drop();
}

PRIMS(fdiv,\x2F) {
   NOS= NOS/TOS;
   _drop();
}

PRIMS(muldiv,\x2A\x2F) {
   long long b,c,d;
   d=TOS;
   c=NOS;
   b=*(int32_t*)SP;
   R0=b*c/d;
   _twodrop();
   TOS=R0;
}

PRIMS(plus1,1+) {
   TOS++;
}

PRIMS(minus1,1-) {
   TOS--;
}

PRIMS(plusstore,+!) {
   *(int32_t*)TOS += NOS;
   _twodrop();
}

PRIMI(branch) { // jump relative
   IP+=*(int32_t*)(IP);
}

PRIMS(nullbranch,0branch) { // jump relative!
   IP+= TOS!=0 ? 4 : *(int32_t*)(IP);
   _drop();
}

PRIMI(immediate) {
   *(char*)(LAST+8)^=IMMED;
}

PRIMS(equal,=) {
   NOS= NOS==TOS;
   _drop();
}

PRIMS(less,<) {
   NOS= NOS<TOS;
   _drop();
}

PRIMS(more,>) {
   NOS= NOS>TOS;
   _drop();
}

PRIMS(iszero,0=) {
   TOS= 0==TOS;
}

PRIMI(negate) {
   TOS= -TOS;
}

PRIMI(begin) { // get actual address
   _pushx(CP);
}

PRIMI(until) {
   *(int32_t*)CP= nullbranch;
   CP+=4;
   *(int32_t*)CP= TOS-CP; // relative distance
   _drop();
   CP+=4;
}

PRIMS(fif,if) {
   *(int32_t*)CP= nullbranch;
   CP+=4;
   _pushx(CP);
   CP+=4; // reserve space
}

PRIMS(fthen,then) {
   *(int32_t*)TOS = CP-TOS; // store distance at address after if
   _drop();
}

PRIMS(felse,else) {
   *(int32_t*)CP= branch;
   CP+=4;
   _pushx(CP);
   CP+=4; // reserve space
   _swap();
   *(int32_t*)TOS = CP-TOS; // store distance at address after if
   _drop();   
}

PRIMI(again) { // begin again endless
   *(int32_t*)CP= branch;
   CP+=4;
   *(int32_t*)CP= TOS-CP; // relative distance
   _drop();
   CP+=4;
}

PRIMS(fwhile,while) {
   *(int32_t*)CP= nullbranch;
   CP+=4;
   _pushx(CP);
   CP+=4; // reserve space
}

PRIMI(repeat) { // begin while repeat
   *(int32_t*)CP= branch;
   CP+=4;
   _swap(); // get address of begin
   *(int32_t*)CP= TOS-CP; // relative distance negative
   _drop();
   CP+=4;
   *(int32_t*)TOS= CP-TOS; // relative distance forward at while
   _drop();
}


PRIMS(fdo,do) { // NOS=end TOS=start
   *(int32_t*)CP=swap; CP+=4; // TOS=end NOS=start
   *(int32_t*)CP=sttora; CP+=4; // end
   *(int32_t*)CP=sttora; CP+=4; // counter
   _pushx(CP); // save loop start address
}

PRIMI(doloop) {
   (*(int32_t*)RP)+=1;
   if(*(int32_t*)RP != *(int32_t*)(RP-4))
   {
      IP+=(*(int32_t*)IP); // jump relative
   }
   else
   {
      RP-=8; // drop counter and end from rack
      IP+=4; // jump over destination
   }
}

PRIMS(floop,loop) {
   *(int32_t*)CP=doloop; CP+=4;
   *(int32_t*)CP=TOS-CP; // relative distance negative
   CP+=4;
   _drop();
}

PRIMI(doplusloop) {
   (*(int32_t*)RP)+=TOS;
   _drop();
   if(*(int32_t*)RP != *(int32_t*)(RP-4))
   {
      IP+=(*(int32_t*)IP); // jump relative
   }
   else
   {
      RP-=8; // drop counter and end from rack
      IP+=4; // jump over destination
   }
}

PRIMS(plusloop,+loop) {
   *(int32_t*)CP=doplusloop; CP+=4;
   *(int32_t*)CP=TOS-CP; // relative distance negative
   CP+=4;
   _drop();
}

PRIMI(i) {
   _pushx(*(int32_t*)RP);
}

PRIMI(j) {
   _pushx(*(int32_t*)(RP-8));
}

PRIMS(ftor,or) {
    NOS|= TOS;
   _drop();
}

PRIMS(ftand,and) {
    NOS&= TOS;
   _drop();
}

PRIMS(ftxor,xor) {
    NOS^= TOS;
   _drop();
}

PRIMS(ftmin,min) {
    NOS= TOS<NOS ? TOS: NOS; 
   _drop();
}

PRIMS(ftmax,max) {
    NOS= TOS>NOS ? TOS: NOS; 
   _drop();
}

PRIMI(mod) {
    NOS= (NOS % TOS);
   _drop();
}

PRIMS(cstore,c!) {
   *(char*)TOS=(char)NOS;
   _twodrop();
}

PRIMS(cat,c@) {
   TOS=*(char*)TOS;
}

PRIMI(fill) { // ( start number char -- )
   R0=TOS;
   _drop();
   R2=TOS+NOS;
   for(R1=NOS;R1<R2;R1++) *(char*)R1=(char)R0;
   _twodrop();
}

// no forth
void compstring(void) { 
   inP++; // skip blank
   while(*(char*)inP!=0 && *(char*)inP!=0x22 ) {
      *(char*)CP=*(char*)inP;
      CP++;
      inP++;
   }
   *(char*)CP++=0;
   while((CP&3)!=0) *(char*)CP++=0; // ALIGNEMENT
   if(*(char*)inP==0x22) inP++; // skip quote
}

PRIMI(getstring) { 
   _pushx(IP+8); // start address of string
}


const char* defstring_s="s\"";
void _defstring(void) { // compile string until NULL or Anfuehrungsz IMMEDIATE
   *(int32_t*)CP= getstring; CP+=4; // compile getstring
   *(int32_t*)CP= branch; CP+=4; // compile branch over string
   R0=CP; 
   CP+=4; // length of string +4 as jump distance
   compstring();
   *(int32_t*)R0=CP-R0; // patch jump distance
}

PRIMS(prints,print$) { // print string from TOS
   printf("%s",(char*)TOS);
   _drop();
}

const char* prtstring_s=".\"";
void _prtstring(void) { // compile string until NULL or Anfuehrungsz IMMEDIATE
   *(int32_t*)CP= getstring; CP+=4; // compile getstring
   *(int32_t*)CP= branch; CP+=4; // compile branch over string
   R0=CP; 
   CP+=4; // length of string +4 as jump distance
   compstring();
   *(int32_t*)R0=CP-R0; // patch jump distance
   *(int32_t*)CP= prints; CP+=4; // compile print$
}

PRIMI(lastfree) { // ( -- varaddr)  last address of free forth memory
   _pushx((int32_t)&lastforth);
}

// Block P2 ------------------------------------------------------------------------
#ifdef P2
FILE *blockfile;
const char blockfilename[]="/sd/blocks.blk";

PRIMI(freadBlock) { // ( textbuf startpos length -- length)
   blockfile= fopen(blockfilename,"rb"); // r Datei muss vorhanden sein, b binary
   R0=TOS; _drop(); // length
   fseek(blockfile,TOS,0);
   R1= fread(NOS, 1, R0, blockfile);
   // buffer, size, count, file
   fclose(blockfile);
   *(char*)(NOS+R0-1)=0; // end of text
   _drop();
   TOS=R1;
}

PRIMI(fwriteBlock) { // ( textbuf startpos length -- )
   blockfile= fopen(blockfilename,"r+b"); // r Datei muss vorhanden sein, b binary
   R0=TOS; _drop(); // length
   fseek(blockfile,TOS,0);
   R1= fwrite(NOS, 1, R0, blockfile);
   // buffer, size, count, file
   fclose(blockfile);
   _twodrop();
}

PRIMI(format) { // dummy
	//blockfile= LittleFS.open(blockfilename,"w"); // write creates file
	//blockfile.close();
}

PRIMI(dir) { // print directory of /sd
   do_dir("/sd");
}

// buffer for holding temporary file names
static char tempname[PATH_MAX];
// ditto for temporary directories
static char tempdir[PATH_MAX];

// P2 utility function: check if a name refers to a directory
// returns 0 if not, 1 if it is
// if an error happens returns 0

int32_t is_directory(const char *filename)
{
    int32_t r;
    struct stat sbuf;
    r = stat(filename, &sbuf);
    if (r) return 0;
    if (S_IFDIR == (sbuf.st_mode & S_IFMT)) {
        return 1;
    }
    return 0;
}

// P2
// perform a directory listing
//
void do_dir(const char *filename)
{
    DIR *d;
    struct dirent *ent;
    struct stat sbuf;
    int32_t r;
    
    if (!filename || filename[0] == 0) {
        getcwd(tempname, sizeof(tempname));
        filename = tempname;
    }
    if (!is_directory(filename)) {
        printf("ERROR: %s is not a directory\n", filename);
        return;
    }
    if ( 0 == (d = opendir(filename)) ) {
        printf("ERROR: unable to read directory %s\n", filename);
        perror(filename);
        return;
    }
    getcwd(tempdir, sizeof(tempdir));
    chdir(filename);
    for(;;) {
        ent = readdir(d); // get next directory entry
        if (!ent) break;  // NULL pointer indicates we are done

        // read information about the file associated with this
        // directory information
        r = stat(ent->d_name, &sbuf);
        if (r) {
            // some kind of error happened
            printf("??? error in stat of %s\n", ent->d_name);
        } else if ((sbuf.st_mode & S_IFMT) == S_IFDIR) {
            // this is a directory
            printf("%8s %s\n", "<dir>", ent->d_name);
        } else {
            // not a directory, print the size
            printf("%8u %s\n", (unsigned)sbuf.st_size, ent->d_name);
        }
    }
    // all done, close the directory
    closedir(d);
    chdir(tempdir);
}

#endif

// Pico Block -----------------------------------------------------------
#ifdef pico
File blockfile;

PRIMI(freadBlock) { // ( textbuf startpos length -- length)
	R0=TOS; _drop(); // length
	char name[20];
	sprintf(name,"block%d",TOS/BSIZE);
	blockfile= LittleFS.open(name,"r"); // read
	if (!blockfile) {
		Serial.println("Failed to open file for reading");
    return;
	}
	if(blockfile) { 

   //blockfile.seek(TOS,SeekSet);
   //printf("\n Seek: %d ", blockfile.seek(TOS,SeekSet));
   R1= blockfile.read((unsigned char*)NOS, R0);
   // buffer, size, count, file
   blockfile.close();
   *(char*)(NOS+R0-1)=0; // end of text
   _drop();
   TOS=R1;
	}
}

PRIMI(fwriteBlock) { // ( textbuf startpos length -- )
	R0=TOS; _drop(); // length
	char name[20];
	sprintf(name,"block%d",TOS/BSIZE);
	blockfile= LittleFS.open(name,"w"); // write
   	if (!blockfile) {
		Serial.println("Failed to open file for writing");
    return; }
   //blockfile.seek(TOS,SeekSet);
   //printf("\n Seek: %d ", blockfile.seek(TOS,SeekSet));
   R1= blockfile.write((unsigned char*)NOS, R0);
   // buffer, size, count, file
   blockfile.close();
   _twodrop(); 
}

PRIMI(format) { // format LittleFS and generate blockfile
	LittleFS.format();
	//blockfile= LittleFS.open(blockfilename,"w"); // write creates file
	//blockfile.close();
}

PRIMI(dir) { // print directory of /sd
/*
   do_dir("/sd");
*/
}
#endif

PRIMI(bSize) { // ( -- value of constant bSize)  
   _pushx((int32_t)BSIZE);
}

PRIMI(bBuf) { // ( -- constant address of bBuf)  
   _pushx((int32_t)(forthMem+FORTHSIZE-BSIZE));
}

PRIMI(actBlock) { // ( -- varaddr)  variable actBlock
   _pushx((int32_t)&factBlock);
}

PRIMI(lineCnt) { // ( -- varaddr)  variable lineCnt
   _pushx((int32_t)&flineCnt);
}

// End Block ------------------------------------------------------------------------


PRIMI(forget) { // wordname
   fword();
   R0=_ffindstr((char*)wordBuf);
   if(R0==0) {
      printf("-- Not found");
      return;
   }
   R1= *(int32_t*)R0; // rewind link
   LAST=R1;
   CONTEXT=R1;
}

PRIMI(context) { // ( -- varaddr)  variable CONTEXT
   _pushx((int32_t)&CONTEXT);
}

PRIMS(finP,inP) { // ( -- varaddr)  variable CONTEXT
   _pushx((int32_t)&inP);
}

PRIMS(ftword,word) { // ( -- addr of wordBuf ) get next word into wordbuf and return address
   fword();
   _pushx((int32_t)wordBuf);
}

PRIMS(interpret,[) { // set interprete state
   status=0;
}

PRIMS(compile,]) { // set compile state
   status=1;
}

PRIMI(depth) { // ( -- depth ) find depth of parameter stack
   _pushx((SP-(UStart+32*4))/4);
}

PRIMI(rdepth) { // ( -- rdepth) find depth of return stack
   _pushx((RP-(UStart+98*4))/4);
}

PRIMS(shiftl,<<) { // ( n1 n2 -- n1<<n2 ) shift left
   NOS= NOS<<TOS;
   _drop();
}

PRIMS(shiftr,>>) { // ( n1 n2 -- n1>>n2 ) shift right
   NOS= (unsigned int32_t)NOS>>TOS;
   _drop();
}

PRIMS(fbackground,background) { // ( -- varaddr)  variable background task
   _pushx((int32_t)&background);
}

PRIMI(pause) { // pause active task and switch to next, if ready
   R0=UStart;  // start at own user area
   if ((int32_t)*(int32_t*)R0==0) return;  // single-tasking
   do{ // look for task that is ready
      R0=(int32_t)*(int32_t*)R0;  // points to next user area
      R1=(int32_t)*(int32_t*)(R0+4); // task status of next task zero for not ready
   } while(R1==0);
   if(R0==UStart) return; // no other task is ready
   // store registers TOS, NOS, SP, RP, auxP, IP, W, EABORT, UStart, status, 
   // toset, R0, R1, R2, R3; // 
   _pushx(0); _pushx(0); // TOS and NOS saved
   *(int32_t*)(UStart+4*2) = EABORT;
   *(int32_t*)(UStart+4*3) = status;
   *(int32_t*)(UStart+4*4) = SP;
   *(int32_t*)(UStart+4*5) = RP;
   *(int32_t*)(UStart+4*6) = auxP;
   *(int32_t*)(UStart+4*7) = IP;
   *(int32_t*)(UStart+4*8) = W;
   
   // get registers from new user area
   UStart=R0;
   EABORT=*(int32_t*)(R0+4*2);
   status=*(int32_t*)(R0+4*3);
   SP=*(int32_t*)(R0+4*4);
   RP=*(int32_t*)(R0+4*5);
   auxP=*(int32_t*)(R0+4*6);
   IP=*(int32_t*)(R0+4*7);
   W=*(int32_t*)(R0+4*8);   
   _twodrop(); // nos and tos retrieved
}

PRIMS(getUser,up@) { // ( -- address ) get user area address
   _pushx(UStart);
}

PRIMS(settrace,trace!) { // ( -- address ) get user area address
   trace=TOS;
   _drop();
}

PRIMS(cmove,cmove) { // ( source dest cnt -- ) move chars starting at source
   R1 = TOS; _drop();
   while(R1>0){
      *(char*)TOS=*(char*)NOS;
      TOS++;
      NOS++;
      R1--;
   }
   _twodrop();
}

PRIMS(rcmove,cmove>) { // ( source dest cnt -- ) move chars starting at source+cnt-1
   R1 = TOS; _drop();
   TOS+=R1-1;
   NOS+=R1-1;
   while(R1>0){
      *(char*)TOS=*(char*)NOS;
      TOS--;
      NOS--;
      R1--;
   }
   _twodrop();
}

// IO Processor specific =======================================================
#ifdef P2 // ------------------------------------------------------------------
const char* cntat_s="cnt@";
void _cntat(void)
{ // ( -- counter )
   _pushx((int32_t)_cnt());
}

void releaseSD( void ){
   umount("/sd");
   _pinh(60); // CS high
   for(int32_t i=0;i<65;i++) {
      _waitx(500);
      _pinnot(61); // CLK
   }
}

const char* bye_s="bye";
void _bye(void) { // reboot
   //printf("\n e to reboot \n");
   //while(getchar()!='e');
   printf("\n Releasing SD \n                   \n");
   releaseSD();
   _waitms(1000);
   printf("\n Rebooting \n                   \n");
   _reboot();
}

PRIMS(pinat,pin@) { // ( pin -- flag )
   TOS=_pinr(TOS);
}

PRIMS(pinstore,pin!) { // ( flag pin -- ) 
   _pinw(TOS, NOS);
   _twodrop();
}

PRIMS(fgetus,getus) { // ( -- us ) 
   R0= (int32_t)_getus();
   _pushx(R0);
}

PRIMS(fgetms,getms) { // ( -- ms ) 
   R0= (int32_t)_getms();
   _pushx(R0);
}

PRIMI(cogstart) { // ( COGnr userArea instruction -- )
   R0= TOS; // instruction
   _drop();
   __builtin_cogstart_cog(NOS,fstartCog(R0,TOS),cstack[NOS]);
   //fstartCog(int32_t instruction, int32_t userPointer) 
   _twodrop();
}
#endif

#ifdef pico // ------------------------------------------------------
const char* cntat_s="cnt@";
void _cntat(void)
{ // ( -- counter )
   //_pushx((int32_t)_cnt());
}

const char* bye_s="bye";
void _bye(void) { // reboot dummy
}

PRIMS(pinat,pin@) { // ( pin -- flag )
   pinMode(TOS,INPUT);
   TOS=digitalRead(TOS);
}

PRIMS(pinstore,pin!) { // ( flag pin -- ) 
   pinMode(TOS,OUTPUT);
   digitalWrite(TOS,NOS);
   _twodrop();
}

PRIMS(fgetus,getus) { // ( -- us ) 
   R0= (int32_t)micros();
   _pushx(R0);
}

PRIMS(fgetms,getms) { // ( -- ms ) 
   R0= (int32_t)millis();
   _pushx(R0);
}

PRIMI(cogstart) { // dummy ( COGnr userArea instruction -- )
}
#endif

// Interpreter #################################################################
#define CAS( code ) case code: _ ## code() ; break
void inner____________dumy() {} // for function list

void inner(void) { // inner interpreter, start code in W
   do{
      if(IP>trace) {
         _prtS4(); // trace
         printf("\n IP: %d Code: %d ", IP, W); 
         fprtWord(W);
      }
      if(W<0) // negative for compound words
      { 
         _rpushx(IP);
         IP=abs(W);
      }
      else { // primitives
            switch(W) {
            case noop : _noop(); break;
            case prtS4 : _prtS4(); break;
            case dump16 : _dump16(); break;
            case drop : _drop(); break;
            case twodrop : _twodrop(); break;
            case dup : _dup(); break;
            case twodup : _twodup(); break;
            case store : _store(); break;
            case at : _at(); break;
            case hereat : _hereat(); break;
            case emit : _emit(); break;
            case dot : _dot(); break;
            case plus : _plus(); break;
            case minus : _minus(); break;
            case wexit : _wexit(); break;
            case lit : _lit(); break;
            case sttora: _sttora(); break;
            case ratost: _ratost(); break;
            case cpratost: _cpratosta(); break;
            case swap: _swap(); break;
            //case fnext: _fnext(); break;
            case rot: _rot(); break;
            case over: _over(); break;
            //case ffor: _sttora(); break; // better read
            case lwords: _lwords(); break;
            CAS(words);
            CAS(findstr);
            CAS(tick);
            CAS(execute);
            CAS(prtword);
            CAS(see);
            CAS(nip);
            CAS(comma);
            CAS(ccomma);
            CAS(colon);
            CAS(semicolon);
            CAS(dofnext);
            CAS(ffor);
            CAS(fnext);
            CAS(evaluate);
            CAS(dovar);
            CAS(variable);
            CAS(docreate); CAS(create);
            CAS(doconst);
            CAS(constant);
            CAS(dovalue); CAS(value); CAS(to); CAS(plusto);
            CAS(toLocA); CAS(plusToLocA); CAS(getLocA); 
            CAS(toLocB); CAS(plusToLocB); CAS(getLocB); 
            CAS(toLocC); CAS(plusToLocC); CAS(getLocC); 
            CAS(toLocD); CAS(plusToLocD); CAS(getLocD); 
            CAS(toLocE); CAS(plusToLocE); CAS(getLocE); 
            CAS(loc); CAS(loc_a); CAS(loc_ab); CAS(loc_abc); CAS(loc_abcd); CAS(loc_abcde); 
            CAS(endloc);
            CAS(backslash);
            CAS(doparen);
            CAS(allot); 
            CAS(mult);
            CAS(fdiv);
            CAS(muldiv);
            CAS(plus1);
            CAS(minus1);
            CAS(plusstore);
            CAS(branch);
            CAS(nullbranch);
            CAS(immediate);
            CAS(equal);
            CAS(less);
            CAS(more);
            CAS(iszero);
            CAS(negate);
            CAS(begin);
            CAS(until);
            CAS(fif);
            CAS(fthen);
            CAS(felse);
            CAS(again);
            CAS(fwhile);
            CAS(repeat);
            CAS(fdo);
            CAS(doloop);
            CAS(floop);
            CAS(doplusloop);
            CAS(plusloop);
            CAS(i);
            CAS(j);
            CAS(ftor);   CAS(ftand);  CAS(ftxor);
            CAS(ftmin);  CAS(ftmax);
            CAS(mod);
            CAS(cstore);
            CAS(cat);
            CAS(fill);
            CAS(getstring);
            CAS(defstring);
            CAS(prints);
            CAS(prtstring);
            CAS(lastfree);
            CAS(freadBlock); 
            CAS(fwriteBlock); 
            CAS(dir);
			CAS(format);
            CAS(bSize); 
            CAS(bBuf); 
            CAS(actBlock); 
            CAS(lineCnt);
            CAS(conkey);
            CAS(gettext);
            CAS(forget);
            CAS(context);
            CAS(finP); 
            CAS(ftword);
            CAS(interpret) ; CAS(compile);
            CAS(depth); CAS(rdepth); CAS(shiftl); CAS(shiftr);
            CAS(fbackground); CAS(pause); CAS(getUser); CAS(settrace);
            CAS(cmove); CAS(rcmove);

            // processor specific
            case cntat: _cntat(); break;
            CAS(bye);
            CAS(pinat); CAS(pinstore);
            CAS(fgetus);
            CAS(fgetms);
            CAS(cogstart);

         }
      }
      if(IP!=0) {
         W=*(int32_t*)IP; 
         IP+=4; 
      }
   } while((IP!=0)&&(EABORT==0));
}


void addName( const char* c ){ // add with length 20 bytes into forthMem
   for(int32_t i=0; i<19 ; i++) {
	   //TOS= LAST; _dump16();
      if(*c != 0) {
        *(char*)CP= *c;
         c++;
        }
      else *(char*)CP= 0; // pad with 0
      CP++;
   }
   *(char*)CP=0;
   CP++;
}

#define COP( code, flags ) *(int32_t*)CP=LAST; LAST=CP; CP+=4; \
	*(int32_t*)CP=0; CP+=4; \
	*(char*)CP= flags; CP+=4; \
   addName(code##_s); *(int32_t*)CP=code; CP+=4

void compPrimitives() { // compile all primitives
	/*
	*(int32_t*)CP=LAST; 
	LAST=CP; CP+=8; 
	*(char*)CP= 0; // flag 
	CP+=1; // 1 !
	printf("\n COP noop 1 %d \n", *(int32_t*)LAST); delay(1000);
   addName(noop_s); 
   printf("\n COP noop 2 \n"); delay(1000);
   *(int32_t*)CP= noop; // hier absturz WEGEN ALIGNEMENT !!!!
   printf("\n COP noop 3 \n"); delay(1000);
   CP+=4;
	*/ 
   COP(noop, 0); 
   //printf("\n COP noop 4 \n"); delay(1000);

   COP(prtS4 , 0);
   COP(dump16 , 0);
   COP(drop, 0);
   COP(twodrop , 0);
   COP(dup , 0);
   COP(twodup,0);
   COP(store , 0);
   COP(at , 0);
   COP(hereat,0);
   COP(emit , 0);
   COP(dot , 0);
   COP(plus , 0);
   COP(minus , 0);
   COP(wexit , 0);
   COP(lit , 0);
   COP(sttora , 0);
   COP(ratost , 0);
   COP(cpratost , 0);
   COP(swap , 0);
   COP(rot , 0);
   COP(over , 0);
   COP(lwords, 0);
   COP(words, 0);
   COP(findstr, 0);
   COP(tick, 0); // IMMED ??
   COP(execute, 0);
   COP(prtword, 0);
   COP(see, 0);
   COP(nip,0);
   COP(comma,0);
   COP(ccomma,0);
   COP(colon,IMMED+ISSTRUCT);
   COP(semicolon,IMMED+ISSTRUCT);
   COP(dofnext,ISLOOP); COP(ffor,IMMED+ISLOOP); COP(fnext,IMMED+ISLOOP);
   COP(evaluate,0);
   
   COP(dovar,0); COP(variable,ISDATA);
   COP(docreate,0); COP(create,ISDATA);
   COP(doconst,0); COP(constant,ISDATA);
   COP(dovalue,0); COP(value,ISDATA); COP(to,0); COP(plusto,0);
   COP(toLocA,0); COP(plusToLocA, 0); COP(getLocA, 0); 
   COP(toLocB,0); COP(plusToLocB, 0); COP(getLocB, 0); 
   COP(toLocC,0); COP(plusToLocC, 0); COP(getLocC, 0); 
   COP(toLocD,0); COP(plusToLocD, 0); COP(getLocD, 0); 
   COP(toLocE,0); COP(plusToLocE, 0); COP(getLocE, 0); 
   COP(loc, 0); COP(loc_a, 0); COP(loc_ab, 0); COP(loc_abc, 0); COP(loc_abcd, 0); COP(loc_abcde, 0); 
   COP(endloc, 0);
   
   COP(backslash,IMMED);
   COP(doparen,IMMED);
   COP(allot,0);
   COP(mult,0);
   COP(fdiv,0);
   COP(muldiv,0);
   COP(plus1,0);
   COP(minus1,0);
   COP(plusstore,0);
   COP(branch,0);
   COP(nullbranch,0);
   COP(immediate,0);
   COP(equal,0);
   COP(less,0);
   COP(more,0);
   COP(iszero,0);
   COP(negate,0);
   COP(begin,IMMED+ISLOOP);
   COP(until,IMMED+ISLOOP);
   COP(fif,IMMED+ISSTRUCT);
   COP(fthen,IMMED+ISSTRUCT);
   COP(felse,IMMED+ISSTRUCT);
   COP(again,IMMED+ISLOOP);
   COP(fwhile,IMMED+ISLOOP);
   COP(repeat,IMMED+ISLOOP);
   COP(fdo,IMMED+ISLOOP);
   COP(doloop,0+ISLOOP);
   COP(floop,IMMED+ISLOOP);
   COP(doplusloop,0+ISLOOP);
   COP(plusloop,IMMED+ISLOOP);
   COP(i,0);
   COP(j,0); 
   COP(ftor,0);    COP(ftand,0);   COP(ftxor,0);
   COP(ftmin,0);   COP(ftmax,0);
   COP(mod,0);
   COP(cstore,0);
   COP(cat,0);
   COP(fill,0);
   COP(getstring,0);
   COP(defstring,IMMED);
   COP(prints,0);
   COP(prtstring,IMMED);
   COP(lastfree,ISDATA);
   COP(freadBlock,ISIO);
   COP(fwriteBlock,ISIO);
   COP(format,ISIO);
   COP(conkey,ISIO);
   COP(gettext,ISIO);
   COP(forget,0);
   COP(dir,0);
   COP(bSize,ISDATA);
   COP(bBuf, ISDATA);
   COP(actBlock, ISDATA);
   COP(lineCnt, ISDATA);
   COP(context, ISDATA);
   COP(finP, ISDATA);
   COP(ftword, 0);
   COP(interpret,IMMED+ISSTRUCT); COP(compile,IMMED+ISSTRUCT);
   COP(depth, 0); COP(rdepth, 0); COP(shiftl,0); COP(shiftr,0);
   COP(fbackground, ISDATA); COP(pause, 0); COP(getUser, ISDATA);
   COP(settrace, ISDATA); 
   COP(cmove, 0); COP(rcmove, 0);

// IO Processor specific
   COP(cntat , 0);   
   COP(bye, 0);
   COP(pinat,ISIO); COP(pinstore, ISIO);
   COP(fgetus,0); COP(fgetms,0);
   COP(cogstart,0);
   // COP( , 0);

   primLimit=LAST;
}

int32_t getnum(char* s ) { //convert a string with hex $ff or decimal
   if((*s>47&&*s<58)||*s==45) { // is a decimal number?
      return atoi(s); 
   }
   s++; // skip $
   char c=*s++;
   R0=0;
   while(c>47) {
      R0=R0*16;
      R0+=(c <= '9') ? c - '0' : ( (c <= 'F') ? c - '7' : c-'7'+65-97);
      c=*s++;
   }
   return R0;
}


void fword( void ) { // copy next token from inP into wordBuf
   *(char*)(wordBuf)=0;
   while(*(char*)inP <33) { // skip leading controls
      if(*(char*)inP ==0) return;
      inP++; 
   }
   for(int32_t i=0;i<19;i++) {
      if(*(char*)inP>32) {
         *(char*)(wordBuf+i)=*(char*)inP++;
      }
      else *(char*)(wordBuf+i)=0;
   }
   *(char*)(wordBuf+19)=0;
   //_pushx(wordBuf); _dump16();
}

void fevalWord(void) { 
   fword();
   if(*(char*)wordBuf==0) return;
   printf(" %s ",(char*)wordBuf);
   R0=_ffindstr((char*)wordBuf); // linkaddr
   if(R0!=0){
      char flags = *(char*)(R0+8);
      if(status==0||(flags&IMMED)){ // interprete
         if(R0<=primLimit) W=*(int32_t*)(R0+32);
         else W=-(R0+32);
         // IP=0; // ??
         inner();
         return;
      }
      else { // compile
         if(R0<=primLimit) *(int32_t*)CP=*(int32_t*)(R0+32);
         else *(int32_t*)CP=-(R0+32);
         CP+=4;
         return;
      }
   }
   if(((*(char*)wordBuf>47&&*(char*)wordBuf<58))||*(char*)wordBuf=='$'||*(char*)wordBuf=='-') { // is a number?
      R1=getnum((char*)wordBuf);
      if(status==0) _pushx(R1);
      else {
         *(int32_t*)CP=lit; CP+=4;
         *(int32_t*)CP=R1; CP+=4;
      }
   }
   else printf(" ??????? \n");
}

// text interpreter
void outerInterp(void) {
   while(1){
      printf("\n RDepth: %d Depth: %d ", (RP-(UStart+98*4))/4, (SP-(UStart+32*4))/4); _prtS4();
      //gets((char*)TIB); // get a line
      bufferedgets((char*)TIB); // get a line
      //_pushx(TIB); _dump16();
      inP=TIB;
      while(*(char*)inP!=0) {
         // fword();
         fevalWord();
         //printf("\n RDepth: %d Depth: %d ", (RP-(int32_t)Rack)/4, (SP-(int32_t)Stack)/4); _prtS4();
      }
   }
}



void builtinWords____________dumy() {} // for function list
const char builtinWords[]= 
   ": fibo ( n -- fibo ) 0 1 rot 1 - ffor swap over + fnext swap drop ; \n" 
   ": fibo46 getus 46 fibo getus swap . swap - . ; \n" 
   "fibo46 \n" 
   ": dump4 ( addr -- ) 4 ffor dup dump16 16 + fnext drop ; \n"
"variable updated 0 updated ! \n"
"27 value maxBlocks# \n"
": flush ( write back block if updated ) \n"
"  updated @ 0= if exit then \n"
"  actBlock @ maxBlocks# < if \n"
"      bBuf   actBlock @ bSize *   bSize fwriteBlock \n"
"	then \n"
"   0 updated ! \n"
"; \n"
": update 1 updated ! ; \n"
": block ( nr -- address ) ( get new block into bBuf) \n"
//"   dup actBlock @ = if drop bBuf exit then \n"
"   flush   dup actBlock ! \n"
"   bSize * ( startpos )   bBuf swap   bSize \n"
"   freadBlock  ( textbuf startpos length -- length) \n"
"   drop bBuf ; \n" 
": load ( block -- )   block drop   bBuf evaluate ; \n"
" 2 load \n"
 ; 
 
void fstartCog(int32_t instruction, int32_t userPointer) { // runs in new COG
   UStart= userPointer;
   setupRegisters();
   W=instruction;
   if(W<=primLimit+29) W=*(int32_t*)W;
   else W=-W;
   inner();
}

void setupRegisters(void) {
   IP=0; TOS=0; NOS=0;
   toset=0; status=0;
   trace=0x70000000; // switch off
   *(int32_t*)UStart=0; // singletask
   *((int32_t*)(UStart+4))=1; // task is ready
   SP= UStart+(32*4); // Parameter Stack
   RP= UStart+(98*4); // Return Stack
   _rpushx(0); // for exit
   auxP= UStart+(162*4);
   EABORT=0;
}

void mymain() {

   lastforth= (int32_t)(forthMem+FORTHSIZE-1);
   TIB=(int32_t)forthMem+0x100;
   wordBuf=(int32_t)forthMem+0x200;
   UStart=(int32_t)forthMem+0x300;
   setupRegisters();
   DictStart=(int32_t)forthMem+0x700;
   CP=(int32_t)forthMem+0x700; 
   CONTEXT=(int32_t)forthMem+0x700; 
   LAST=0;
   printf("\n Mem %x User %x Dict %x \n", (int)forthMem, UStart, CP); 
   
   compPrimitives();
   CONTEXT=LAST;

   // _words();
   
   _pushx((int32_t)builtinWords);   _evaluate();
   
   printf("\n %s \n", Titel);
   _words();
   
   outerInterp();
   printf("\n END \n");
   _bye();
}

#ifdef pico
void setup() {
  // put your setup code here, to run once:
  Serial.begin(115200);
  delay(5000);
  printf("\n Started \n");
  Serial.println(__FILE__);
  Serial.println(__DATE__);
  Serial.println(__TIME__);
  
	LittleFS.begin();
   forthMem= (char*)malloc(FORTHSIZE);
  
  mymain();
}

void loop() {
  // put your main code here, to run repeatedly:
  delay(10000);
}
#endif

#ifdef P2
void main(void) {
   __builtin_cogstart_cog(1,bufferedInput(),bistack); // start buffered serial input

   R0=mount("/sd", _vfs_open_sdcard());
   if (R0 == 0) {
        printf("mounted SD card as /sd\n");
    }
   do_dir("/sd");

	mymain();
}
#endif
